const io = @import("./io.zig");
const testing = @import("./testing.zig");

///
/// Returns a fixed-size stack type of `Element`s.
///
pub fn Fixed(comptime Element: type) type {
    return struct {
        filled: usize = 0,
        buffer: []Element,

        ///
        /// Stack type.
        ///
        const Self = @This();

        ///
        /// Resets the number of filled items to `0`, otherwise leaving the actual memory contents
        /// of the buffer untouched until it is later overwritten by following operations on it.
        ///
        pub fn clear(self: *Self) void {
            self.filled = 0;
        }

        ///
        /// If `self` is filled with at least `1` value, it is decremented by `1`, otherwise leaving
        /// the actual memory contents of the buffer untouched until it is later overwritten by
        /// following operations on it.
        ///
        /// The value of the element removed from the list is returned if something existed to be
        /// popped, otherwise `null` if it contained no elements.
        ///
        pub fn pop(self: *Self) ?Element {
            if (self.filled == 0) return null;

            self.filled -= 1;

            return self.buffer[self.filled];
        }

        ///
        /// Attempts to push `element` into `self`, returning a [FixedPushError] if it failed.
        ///
        pub fn push(self: *Self, element: Element) PushError!void {
            if (self.filled == self.buffer.len) return error.OutOfMemory;

            self.buffer[self.filled] = element;
            self.filled += 1;
        }

        ///
        /// Attempts to push all of `elements` into `self`, returning a [FixedPushError] if it
        /// failed.
        ///
        pub fn pushAll(self: *Self, elements: []const Element) PushError!void {
            const filled = (self.filled + elements.len);

            if (filled > self.buffer.len) return error.OutOfMemory;

            io.copy(Element, self.buffer[self.filled ..], elements);

            self.filled = filled;
        }
    };
}

test "Fixed stack of string literals" {
    const default_value = "";
    var buffer = [_][]const u8{default_value} ** 4;
    var shopping_list = Fixed([]const u8){.buffer = &buffer};

    // Pop empty stack.
    {
        try testing.expect(shopping_list.pop() == null);
        try testing.expect(shopping_list.filled == 0);
        try testing.expect(shopping_list.buffer.ptr == &buffer);
        try testing.expect(shopping_list.buffer.len == buffer.len);

        for (shopping_list.buffer) |item|
            try testing.expect(io.equals(u8, item, default_value));
    }

    // Push single element.
    {
        try shopping_list.push("milk");
        try testing.expect(shopping_list.filled == 1);
        try testing.expect(shopping_list.buffer.ptr == &buffer);
        try testing.expect(shopping_list.buffer.len == buffer.len);
        try testing.expect(io.equals(u8, shopping_list.buffer[0], "milk"));

        for (shopping_list.buffer[1 ..]) |item|
            try testing.expect(io.equals(u8, item, default_value));

        // TODO: Test stack overflow.
    }

    // Pop single element.
    {
        try testing.expect(io.equals(u8, shopping_list.pop().?, "milk"));
        try testing.expect(shopping_list.filled == 0);
        try testing.expect(shopping_list.buffer.ptr == &buffer);
        try testing.expect(shopping_list.buffer.len == buffer.len);
        try testing.expect(io.equals(u8, shopping_list.buffer[0], "milk"));

        for (shopping_list.buffer[1 ..]) |item|
            try testing.expect(io.equals(u8, item, default_value));
    }

    // TODO: Multiple elements.
    // TODO: Clear elements.
}

///
/// Potential errors that may occur while trying to push one or more elements into a stack.
///
pub const PushError = io.Allocator.MakeError;

///
/// Returns an [io.Writer] wrapping `fixed_stack`.
///
/// Writing to the returned [io.Writer] will push values to the underlying [Fixed] stack instance
/// referenced by `fixed_stack` until it is full.
///
pub fn fixedWriter(fixed_stack: *Fixed(u8)) io.Writer {
    return io.Writer.capture(fixed_stack, struct {
        fn write(stack: *Fixed(u8), buffer: []const u8) usize {
            stack.pushAll(buffer) catch |err| switch (err) {
                error.OutOfMemory => return 0,
            };

            return buffer.len;
        }
    }.write);
}

test "Fixed writer" {
    var buffer = [_]u8{0} ** 4;
    var sequence_stack = Fixed(u8){.buffer = &buffer};
    const sequence_data = [_]u8{8, 16, 32, 64};

    try testing.expect(fixedWriter(&sequence_stack).call(&sequence_data) == sequence_data.len);
    try testing.expect(io.equals(u8, sequence_stack.buffer, &sequence_data));
}
